#<profile4xi profile = "10" />
# -------------------------------------------------------------------------------------------------
#<program4xi program = "wait4xi.py">
#<abstract4xi abstract = "provides the function wait4xi"/>
#<refer4xi reference = "connection.py"/>
#<note4xi note = "called from several platforms - platform dependent"/>
#<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
# -------------------------------------------------------------------------------------------------

# -------------------------------------------------------------------------------------------------
# Public API
# -------------------------------------------------------------------------------------------------
#<interface4xi interface = " 
__all__ = ['wait4xi']
#"
#/>


# -------------------------------------------------------------------------------------------------
# Standard Imports                          
# -------------------------------------------------------------------------------------------------
#<standardize4xi standard = "
import errno
import io
import itertools
import os
import pickle
import select
import socket
import struct
import sys
import threading                
import _thread
import tempfile
import time
try:
    import _winapi
    from _winapi import WAIT_OBJECT_0, WAIT_TIMEOUT, INFINITE
except ImportError:
    if sys.platform == 'win32':
        raise
    _winapi = None


#"
#/>

# -------------------------------------------------------------------------------------------------
# Multiprocessing Imports                          
# -------------------------------------------------------------------------------------------------
#<process4xi process = "
import _multiprocessing
from multiprocessing import connection
#"
#/>

# -------------------------------------------------------------------------------------------------
# DragonXi imports                              
# -------------------------------------------------------------------------------------------------
#<dragonize4xi dragonization = "
from print4xi import print4xi
from time4xi import time4xi
from timeout4xi import omega2timeout4xi
from _exhaustive_wait4xi import _exhaustive_wait4xi
#"
#/>

# -------------------------------------------------------------------------------------------------
# Settings
# -------------------------------------------------------------------------------------------------
#<set4xi setting = "
BUFSIZE = 8192
CONNECTION_TIMEOUT = 20.
_mmap_counter = itertools.count()
default_family = 'AF_INET'
families = ['AF_INET']

if hasattr(socket, 'AF_UNIX'):
    default_family = 'AF_UNIX'
    families += ['AF_UNIX']

if sys.platform == 'win32':
    default_family = 'AF_PIPE'
    families += ['AF_PIPE']
_ready_errors = {_winapi.ERROR_BROKEN_PIPE, _winapi.ERROR_NETNAME_DELETED}
#"
#"/>



if sys.platform == 'win32':
    # -------------------------------------------------------------------------------------------------
    #<function4xi function = "wait4xi" usage = "
    #   connection.py
    #       class Connection(_ConnectionBase):
    #           def _poll(self, timeout):
    #               r = wait([self._handle], timeout)
    #
    #   connection4xi.py
    #       class Connection4Xi(_ConnectionBase):
    #           def _poll(self, timeout):
    #               r = wait4xi([self._handle], timeout)
    #   pipeconnection4xi.py
    #       class PipeConnection4Xi(_ConnectionBase4Xi):
    #           def _poll(self, timeout):
    #               ...
    #               return bool(wait4xi([self], timeout))
    #    
    #"> 
    #<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
    # -------------------------------------------------------------------------------------------------
    def wait4xi(object_list, timeout=None):
        '''
        Wait till an object in object_list is ready/readable.
        Return list of those objects in object_list which are ready/readable.
        '''
        
        # -----------------------------------------------------------------------------------------    
        # (Step 0/5) 
        # -----------------------------------------------------------------------------------------        
        class2print = "wait4xi"
        method2print = "wait4xi"
    
        
        message2print = "{(Step 0a/5)  started" + "; wait4xi-started: " + str(wait4xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
        #     
        
        message2print = "{(Step 0b/5) passed" + "; object_list-passed-to-wait4xi: " + str(object_list) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        # 
         
        message2print = "{(Step 0c/5) passed" + "; timeout-passed-to-wait4xi: " + str(timeout) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        # 
                    
        # -----------------------------------------------------------------------------------------    
        # (Step 1/5) 
        # -----------------------------------------------------------------------------------------        
        if timeout is None:
            timeout = omega2timeout4xi(class2print, method2print)                                                   # dragonized INFINITE
            message2print = "{(Step 1/5a) set" + "; timeout-set-by-wait4xi to: " + str(timeout)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            #   
        elif timeout < 0:
            timeout = 0
            message2print = "{(Step 1/5b) set" + "; timeout-set-by-wait4xi to: " + str(timeout)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            #   
        else:
            timeout = int(timeout * 1000 + 0.5)
            message2print = "{(Step 1/5c) set" + "; timeout-set-by-wait4xi to: " + str(timeout)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            #   
            
            max2timeout = omega2timeout4xi(class2print, method2print)
            
            if timeout > max2timeout:                                                                               # dragonized not to exceed maximum
                timeout = max2timeout
                message2print = "{(Step 1/5d) set" + "; timeout-set-by-wait4xi to: " + str(timeout)  + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)
                # 
                    
        # -----------------------------------------------------------------------------------------    
        # (Step 2/5) 
        # -----------------------------------------------------------------------------------------           
        object_list2xi = list(object_list)
        message2print = "{(Step 2/5a) set" + "; object_list2xi-set-by-wait4xi to: " + str(object_list2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #   
        waithandle_to_obj2xi = {}
        message2print = "{(Step 2/5a) set" + "; waithandle_to_obj2xi-set-by-wait4xi to: " + str(waithandle_to_obj2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #  
        ov_list2xi = []
        message2print = "{(Step 2/5a) set" + "; ov_list2xi-set-by-wait4xi to: " + str(ov_list2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #  
        ready_objects2xi = set()
        message2print = "{(Step 2/5a) set" + "; ready_objects2xi-set-by-wait4xi to: " + str(ready_objects2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #  
        ready_handles2xi = set()
        message2print = "{(Step 2/5a) set" + "; ready_handles2xi-set-by-wait4xi to: " + str(ready_handles2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #  
                    
        # -----------------------------------------------------------------------------------------    
        # (Step 3/5) 
        # -----------------------------------------------------------------------------------------   
        o2xi = None
        try:
            for o2xi in object_list2xi:
                message2print = "{(Step 3/5a) using" + "; o2xi-using-by-wait4xi: " + str(o2xi)  + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)
                #  
                try:
                    fileno2xi = getattr(o2xi, 'fileno')
                    message2print = "{(Step 3/5b) got" + "; fileno2xi-got-by-wait4xi: " + str(fileno2xi)  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)
                    # 
                    
                except AttributeError:
                    waithandle_to_obj2xi[o2xi.__index__()] = o2xi
                    message2print = "{(Step 3/5c) set" + "; waithandle-set-by-wait4xi to: " + str(waithandle_to_obj2xi[o2xi.__index__()])  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)
                    # 
                else:
                    # start an overlapped read of length zero
                    try:
                        ov2xi, err2xi = _winapi.ReadFile(fileno2xi(), 0, True)
                        message2print = "{(Step 3/5d) got" + "; ov2xi-got-by-wait4xi: " + str(ov2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        # 
                        message2print = "{(Step 3/5e) got" + "; err2xi-got-by-wait4xi: " + str(err2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        # 
                    except OSError as e2xi:
                        err2xi = e2xi.winerror
                        message2print = "{(Step 3/5z1) got" + "; err2xi-got-by-wait4xi: " + str(err2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        # 
                        if err2xi not in _ready_errors:
                            message2print = "{(Step 3/5z2) failed" + "; err2xi-failed-by-wait4xi: " + str(err2xi)  + "; time: " + time4xi() + "}"
                            print4xi(class2print, method2print, message2print)
                            # 
                            raise
                        else:
                            message2print = "{(Step 3/5z3) warned" + "; err2xi-warned-by-wait4xi: " + str(err2xi)  + "; time: " + time4xi() + "}"
                            print4xi(class2print, method2print, message2print)
                            # 
                                
                    if err2xi == _winapi.ERROR_IO_PENDING:
                        message2print = "{(Step 3/5f) got" + "; err2xi-got-by-wait4xi: " + str(err2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        # 
                        ov_list2xi.append(ov2xi)
                        
                        message2print = "{(Step 3/5g) appended" + "; waithandle-appended-by-wait4xi: " + str(ov2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        # 
                        
                        waithandle_to_obj2xi[ov2xi.event] = o2xi
 
                        message2print = "{(Step 3/5h) using" + "; o2xi-using-by-wait4xi: " + str(o2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        #                                                
                        message2print = "{(Step 3/5i) set" + "; waithandle-set-by-wait4xi to: " + str(waithandle_to_obj2xi[ov2xi.event])  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        # 
                    else:
                        # If o2xi.fileno2xi() is an overlapped pipe handle and
                        # err2xi == 0 then there is a zero length message
                        # in the pipe, but it HAS NOT been consumed.
                        message2print = "{(Step 3/5j) using" + "; o2xi-using-by-wait4xi: " + str(o2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        #    
                        ready_objects2xi.add(o2xi)
                        
                        message2print = "{(Step 3/5k) updated" + "; ready_objects2xi-updated-by-wait4xi: " + str(ready_objects2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        #   
                        
                        timeout = 0
                        message2print = "{(Step 3/5l) set" + "; timeout-set-by-wait4xi to: " + str(timeout)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        # 

            message2print = "{(Step 3/5m) using" + "; keys2objects-using-by-wait4xi: " + str(waithandle_to_obj2xi.keys())  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            #                           
                        
            ready_handles2xi = _exhaustive_wait4xi(waithandle_to_obj2xi.keys(), timeout)
            message2print = "{(Step 3/5n) set" + "; ready_handles2xi-set-by-wait4xi to: " + str(ready_handles2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            #  
        except:
            message2print = "{(Step 3/5z) failed" + "; o2xi-failed-by-wait4xi: " + str(o2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            #   
                       
        # -----------------------------------------------------------------------------------------    
        # (Step 4/5) 
        # -----------------------------------------------------------------------------------------               
        finally:
            # request that overlapped reads stop
            for ov2xi in ov_list2xi:
                message2print = "{(Step 4/5a) using" + "; ov2xi-using-by-wait4xi: " + str(ov2xi)  + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)
                #   
                ov2xi.cancel()
                message2print = "{(Step 4/5b) performed" + "; cancel-performed-by-wait4xi: " + str(ov2xi.cancel)  + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)
                #   

            # wait for all overlapped reads to stop
            message2print = "{(Step 4/5c) using" + "; ov_list2xi-using-by-wait4xi: " + str(ov_list2xi)  + "; time: " + time4xi() + "}"
            print4xi(class2print, method2print, message2print)
            #  
            for ov2xi in ov_list2xi:
                message2print = "{(Step 4/5d) using" + "; ov2xi-using-by-wait4xi: " + str(ov2xi)  + "; time: " + time4xi() + "}"
                print4xi(class2print, method2print, message2print)
                #  
                try:
                    _, err2xi = ov2xi.GetOverlappedResult(True)
                    message2print = "{(Step 4/5e) got" + "; err2xi-got-by-wait4xi: " + str(err2xi)  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)
                    #  
                except OSError as e2xi:
                    err2xi = e2xi.winerror
                    message2print = "{(Step 4/5z1) got" + "; err2xi-got-by-wait4xi: " + str(err2xi)  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)
                    #  
                    if err2xi not in _ready_errors:
                        message2print = "{(Step 4/5z2) failed" + "; err2xi-failed-by-wait4xi: " + str(err2xi)  + "; time: " + time4xi() + "}"
                        print4xi(class2print, method2print, message2print)
                        #  
                        raise
                if err2xi != _winapi.ERROR_OPERATION_ABORTED:
                    o2xi = waithandle_to_obj2xi[ov2xi.event]
                    message2print = "{(Step 4/5f) got" + "; o2xi-got-by-wait4xi: " + str(o2xi)  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)
                    #  
                    ready_objects2xi.add(o2xi)
                    message2print = "{(Step 4/5g) added" + "; ready_objects2xi-added-by-wait4xi: " + str(ready_objects2xi)  + "; time: " + time4xi() + "}"
                    print4xi(class2print, method2print, message2print)
                    #  
                    
                    if err2xi == 0:
                        # If o2xi.fileno2xi() is an overlapped pipe handle then
                        # a zero length message HAS been consumed.
                        if hasattr(o2xi, '_got_empty_message'):
                            o2xi._got_empty_message = True
                            message2print = "{(Step 4/5h) set" + "; got2empty-set-by-wait4xi to: " + str(o2xi._got_empty_message)  + "; time: " + time4xi() + "}"
                            print4xi(class2print, method2print, message2print)
                            #  

        message2print = "{(Step 4/5i) using" + "; ready_handles2xi-using-by-wait4xi: " + str(ready_handles2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #                      
        ready_objects2xi.update(waithandle_to_obj2xi[h2xi] for h2xi in ready_handles2xi)
        
        
        message2print = "{(Step 4/5j) updated" + "; ready_objects2xi-updated-by-wait4xi: " + str(ready_objects2xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)
        #                    
            
           
        # -----------------------------------------------------------------------------------------    
        # (Step 5/5)
        # -----------------------------------------------------------------------------------------     
        message2print = "{(Step 5/5) returning" + "; readyobjects-returning-by-wait4xi: " + str([o2xi for o2xi in object_list2xi if o2xi in ready_objects2xi]) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
        # 
        
        return [o2xi for o2xi in object_list2xi if o2xi in ready_objects2xi]
    #</function4xi>
        
else:
    # -------------------------------------------------------------------------------------------------
    #<function4xi function = "wait4xi" usage = "
    #   connection.py
    #       class Connection(_ConnectionBase):
    #           def _poll(self, timeout):
    #               r = wait([self._handle], timeout)
    #
    #   connection4xi.py
    #       class Connection4Xi(_ConnectionBase):
    #           def _poll(self, timeout):
    #               r = wait4xi([self._handle], timeout)
    #    
    #"> 
    #<state4xi status = "ACK4tagging; ACK4moduletesting; NACK4systemtesting; [Rel 7.21]"/>
    # -------------------------------------------------------------------------------------------------
    def wait4xi(object_list, timeout=None):
        '''
        Wait till an object in object_list is ready/readable.
        Return list of those objects in object_list which are ready/readable.
        '''
        
        # -----------------------------------------------------------------------------------------    
        # (Step 1/3) 
        # -----------------------------------------------------------------------------------------        
        class2print = "wait4xi"
        method2print = "wait4xi"
    
        
        message2print = "{(Step 1a/3)  started" + "; wait4xi-started: " + str(wait4xi)  + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print) 
        #     
        
        message2print = "{(Step 1b/3) passed" + "; object_list-passed-to-wait4xi: " + str(object_list) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        # 
         
        message2print = "{(Step 1c/3) passed" + "; timeout-passed-to-wait4xi: " + str(timeout) + "; time: " + time4xi() + "}"
        print4xi(class2print, method2print, message2print)  
        # 
                    
        # -----------------------------------------------------------------------------------------    
        # (Step 2/3)
        # -----------------------------------------------------------------------------------------              
        if timeout is not None:
            if timeout <= 0:
                message2print = "{(Step 2/3a) returning" + "; object_list-returning-by-wait4xi: " + str(object_list) + "; time: " + time4xi() + "}" 
                print4xi(class2print, method2print, message2print)    
                # 
                 
                return select.select(object_list, [], [], 0)[0]
            else:
                deadline2xi = time.time() + timeout                    # time.time() = 1 353 072 551.2244 at 5:30AM 
                message2print = "{(Step 2/3b) set" + "; deadline2xi-set-by-wait4xi to: " + str(deadline2xi) + "; time: " + time4xi() + "}" 
                print4xi(class2print, method2print, message2print)    
                # 
        else:
            pass         
                    
        # -----------------------------------------------------------------------------------------    
        # (Step 3/3) 
        # -----------------------------------------------------------------------------------------                    
        message2print = "{(Step 3/3a) using" + "; timeout-using-by-wait4xi: " + str(timeout) + "; time: " + time4xi() + "}" 
        print4xi(class2print, method2print, message2print)    
        # 
        while True:
            try:
                message2print = "{(Step 3/3b) returning" + "; object_list-returning-by-wait4xi: " + str(object_list) + "; time: " + time4xi() + "}" 
                print4xi(class2print, method2print, message2print)    
                # 
                return select.select(object_list, [], [], timeout)[0]
            except OSError as e2xi:
                if e2xi.errno != errno.EINTR:
                    message2print = "{(Step 3/3z1) failed" + "; errno-failed-by-wait4xi: " + str(e2xi.errno) + "; time: " + time4xi() + "}" 
                    print4xi(class2print, method2print, message2print)    
                    # 
                    raise
            if timeout is not None:
                timeout = deadline2xi - time.time()
                message2print = "{(Step 3/3c) set" + "; timeout-set-by-wait4xi to: " + str(timeout) + "; time: " + time4xi() + "}" 
                print4xi(class2print, method2print, message2print)    
                # 
    #</function4xi>
#</program4xi>

# -------------------------------------------------------------------------------------------------
# Recorded History                             
# ------------------------------------------------------------------------------------------------- 
#<record4xi record = "
#   testwait4xi.txt
#   sense-test-wait4xi.xml
#   reason-test-wait4xi.xml
#   ace-test-wait4xi.xml
#"
#>


# -------------------------------------------------------------------------------------------------
# Legal Notice                             
# ------------------------------------------------------------------------------------------------- 
#<protect4xi protection = "copyright"
#   owner = "Small & Smart"
#   rights = "All rights reserved"
#   state = "California"
#   country = "United States of America"
#   references = "Legal notice at www.dragonxi.com; www.yourdragonxi.com; www.yourdragonxi.org; www.dragon2xi.com/; www.dragon4xi.com/; www.dragon8xi.com/ " 
#   codes = "Codes established in California and applicable to corporations in California"
#   compliance = "The California Law"
#   court = "Supreme Court of California, 350 McAllister Street, San Francisco, CA 94102-4783"
#/>
 

#<end4xi end = "
# -------------------------------------------------------------------------------------------------
# end of wait4xi.py                             
# ------------------------------------------------------------------------------------------------- 
#"
#/>


